"""Neural Network-Boosted Importance Sampling for Bayesian Statistics."""

from .sampler import Sampler
from .prior import Prior

__author__ = 'Johannes U. Lange'
__email__ = 'jlange@american.edu'
__version__ = '1.0.6'
__all__ = ['Prior', 'Sampler']
